signature Type =
sig

  type t
  val mk_vartype : string -> t
  val mk_type : string -> t list -> t
    (* can raise exceptions,
       e.g., if I call mk_type("fun", [])
    *)
  val arity : string -> int
end


structure Type :> Type =
struct

(*
   data TY = Vartype String |
             Tyop (String,[TY])
*)

(* need to implement a Σₜ that enforces
   arities for existing types, and doesn't
   permit unknown types to occur
*)

datatype t = Vartype of string
           | Tyop of string * t list;



end
open Type
val num_ty = Tyop("num", []);;

val type_of_map =
    Tyop("fun", [
      Tyop("fun", [Vartype "'a",
                   Vartype "'b"]),
      Tyop("fun",
           [Tyop("list", [Vartype "'a"]),
            Tyop("list", [Vartype "'b"])])])


val alpha = mk_vartype ("'a")
val beta = mk_vartype "'b"
infixr -->
fun (a --> b) = Tyop ("fun", [a,b])
fun list_ty a = Tyop ("list", [a])

signature Term =
sig

  type t
  val mk_var : string * Type.t -> t
  val mk_con : string * Type.t -> t (* EXN *)
  val mk_app : t * t -> t  (* EXN *)
  val mk_lam : string * Type.t * t -> t

  val sign : (* dictionary *)

  val aeq : t * t -> bool
  val subst : t * t -> t
  val typeof : t -> Type.t
end

fun mk_eq(t1,t2) =
  (* check t1 and t2 have same type τ *)
  App(App(Con ("=", τ --> τ --> bool),
          t1),
      t2)


structure Term = (* implementation *)
struct
end

structure Thm =
struct

  datatype thm =
    term list * term (* must be of type bool *)

end


signature Thm =
sig

  type thm
  type term = Term.t
  val ASSUME : term -> thm

    (* fun ASSUME t = (* check it's bool *)
            ([t], t) *)

  val REFL : term -> t
    (* fun REFL t =
         ([], mk_eq(t,t))

     *)

  val APPEQ : thm -> thm -> thm
   (* Γ ⊢ f = g , Δ ⊢ x = y;
        return Γ ∪ Δ ⊢ f x = g y
     check both theorems are equalities, and
     that f can be applied to x *)

  val ABSEQ : string * Type.t -> thm -> thm
   (* check i. var(s,ty) ∉ assumptions of thm;
           ii. thm is an equality
   *)

  val BETA : term -> thm
   (* check tm is abstraction applied to arg *)

   (λx. M) N = M[x:=N]
   (* many issues arising;
      rename embedded bound variables if
      necessary to avoid capture
 *)

   (λx y. x y) y = (λy'. y y')


  val EQ_MP : thm -> thm -> thm
    (* ⊢ x = y, ⊢ x gives ⊢ y *)
  val EQ_ANTISYM: (* analog of x ⇒ y, y ⇒ x
                     gives x ⇔ y *)
     thm -> thm -> thm

  val INST_TYPE : tysubst -> thm -> thm
  val INST : tmsubst -> thm -> thm

  val new_constdef:
    string * Type.t * term -> thm
   (* side effect: extend signature with
      new constant of given name and type *)
   (* i. check type of term = given type;
      ii. no mention of var(s,τ) in term
      iii. constant shouldn't be in signature
      iv. tyvars on right, must be in input ty
   *)

  val new_axiom : term -> thm
    (* fun new_axiom t = ([], t) *)
end


end
